"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const vscode_1 = require("vscode");
function inputModeEnabled() {
    return vscode_1.workspace.getConfiguration('lean.input').get('enabled', true);
}
function inputModeLeader() {
    return vscode_1.workspace.getConfiguration('lean.input').get('leader', '\\');
}
function inputModeLanguages() {
    return vscode_1.workspace.getConfiguration('lean.input').get('languages', ['lean']);
}
exports.inputModeLanguages = inputModeLanguages;
function inputModeCustomTranslations() {
    return vscode_1.workspace.getConfiguration('lean.input').get('customTranslations', {});
}
/** Adds hover behaviour for getting translations of unicode characters. Eg: "Type ⊓ using \glb or \sqcap"  */
class LeanInputExplanationHover {
    constructor(translations) {
        this.translations = translations;
        this.leader = inputModeLeader();
        this.customTranslations = inputModeCustomTranslations();
        this.subscriptions = [];
        this.subscriptions.push(vscode_1.workspace.onDidChangeConfiguration(() => {
            this.leader = inputModeLeader();
            this.customTranslations = inputModeCustomTranslations();
        }));
    }
    getAbbrevations(symbol) {
        const abbrevs = [];
        for (const k in this.customTranslations) {
            if (this.customTranslations[k] === symbol) {
                abbrevs.push(k);
            }
        }
        for (const k in this.translations) {
            if (this.customTranslations[k]) {
                continue;
            }
            if (this.translations[k] === symbol) {
                abbrevs.push(k);
            }
        }
        return abbrevs;
    }
    provideHover(document, pos, token) {
        const symbolRange = new vscode_1.Range(pos, pos.translate(0, 1));
        const symbol = document.getText(symbolRange);
        const abbrevs = this.getAbbrevations(symbol).sort((a, b) => a.length - b.length);
        return abbrevs.length > 0 &&
            new vscode_1.Hover(`Type ${symbol} using ${abbrevs.map((a) => this.leader + a).join(' or ')}`, symbolRange);
    }
    dispose() {
        for (const s of this.subscriptions) {
            s.dispose();
        }
    }
}
exports.LeanInputExplanationHover = LeanInputExplanationHover;
/* Each editor has their own abbreviation handler. */
class TextEditorAbbrevHandler {
    constructor(editor, abbreviator) {
        this.editor = editor;
        this.abbreviator = abbreviator;
    }
    updateRange(range) {
        if (range && !range.isSingleLine) {
            range = null;
        }
        this.range = range;
        this.editor.setDecorations(this.abbreviator.decorationType, range ? [range] : []);
        this.abbreviator.updateInputActive();
        // HACK: support \{{}} and \[[]]
        const hackyReplacements = {
            [this.leader + '{{}}']: '⦃⦄',
            [this.leader + '[[]]']: '⟦⟧',
            [this.leader + '<>']: '⟨⟩',
            [this.leader + '([])']: '⟮⟯',
            [this.leader + 'f<>']: '‹›',
        };
        if (range) {
            const replacement = hackyReplacements[this.editor.document.getText(range)];
            if (replacement) {
                this.editor.edit((builder) => __awaiter(this, void 0, void 0, function* () {
                    yield builder.replace(range, replacement);
                    const pos = range.start.translate(0, 1);
                    this.editor.selection = new vscode_1.Selection(pos, pos);
                    this.updateRange();
                }));
            }
        }
    }
    get leader() { return this.abbreviator.leader; }
    get enabled() { return this.abbreviator.enabled; }
    get rangeSize() {
        return this.range.end.character - this.range.start.character;
    }
    convertRange(newRange) {
        if (!this.range || this.rangeSize < 2) {
            return this.updateRange();
        }
        const range = this.range;
        const toReplace = this.editor.document.getText(range);
        if (toReplace[0] !== this.leader) {
            return this.updateRange();
        }
        const abbreviation = toReplace.slice(1);
        const replacement = this.abbreviator.findReplacement(abbreviation);
        if (replacement) {
            setTimeout(() => __awaiter(this, void 0, void 0, function* () {
                // Without the timeout hack, inserting `\delta ` at the beginning of an
                // existing line would leave the cursor four characters too far right.
                yield this.editor.edit((builder) => builder.replace(range, replacement));
                if (newRange) {
                    this.updateRange(new vscode_1.Range(newRange.start.translate(0, replacement.length - toReplace.length), newRange.end.translate(0, replacement.length - toReplace.length)));
                }
            }), 0);
        }
        this.updateRange(newRange);
    }
    onChanged(ev) {
        if (ev.contentChanges.length === 0) {
            // This event is triggered by files.autoSave=onDelay
            return;
        }
        if (ev.contentChanges.length !== 1) {
            return this.updateRange();
        } // single change
        const change = ev.contentChanges[0];
        if (change.text.length === 1 || change.text === '\r\n') {
            // insert (or right paren overwriting)
            if (!this.range) {
                if (change.text === this.leader) {
                    return this.updateRange(new vscode_1.Range(change.range.start, change.range.start.translate(0, 1)));
                }
            }
            else if (change.range.start.isEqual(this.range.end)) {
                if (change.text === this.leader && this.rangeSize === 1) {
                    this.updateRange();
                    return this.editor.edit((builder) => builder.delete(new vscode_1.Range(change.range.start, change.range.end.translate(0, 1))));
                }
                else if (change.text === this.leader) {
                    return this.convertRange(new vscode_1.Range(change.range.start, change.range.start.translate(0, 1)));
                }
                else if (change.text.match(/^\s+|[)}⟩]$/)) {
                    // whitespace, closing parens
                    return this.convertRange();
                }
            }
        }
        if (this.range && this.range.contains(change.range) && this.range.start.isBefore(change.range.start)) {
            // modification
            return this.updateRange(new vscode_1.Range(this.range.start, this.range.end.translate(0, change.text.length - change.rangeLength)));
        }
        this.updateRange();
    }
    onSelectionChanged(ev) {
        if (ev.selections.length !== 1 || !this.range.contains(ev.selections[0].active)) {
            this.convertRange();
        }
    }
}
class LeanInputAbbreviator {
    constructor(translations) {
        this.translations = translations;
        this.subscriptions = [];
        this.leader = inputModeLeader();
        this.enabled = inputModeEnabled();
        this.languages = inputModeLanguages();
        this.customTranslations = inputModeCustomTranslations();
        this.handlers = new Map();
        this.translations = Object.assign({}, translations);
        this.allTranslations = Object.assign(Object.assign({}, this.translations), this.customTranslations);
        this.decorationType = vscode_1.window.createTextEditorDecorationType({
            textDecoration: 'underline',
        });
        this.subscriptions.push(vscode_1.workspace.onDidChangeTextDocument((ev) => this.onChanged(ev)));
        this.subscriptions.push(vscode_1.window.onDidChangeTextEditorSelection((ev) => this.onSelectionChanged(ev)));
        this.subscriptions.push(vscode_1.window.onDidChangeVisibleTextEditors((editors) => {
            // delete removed editors
            const handlers = new Map();
            this.handlers.forEach((h, e) => {
                if (editors.indexOf(e) !== -1) {
                    handlers.set(e, h);
                }
            });
            this.handlers = handlers;
        }));
        this.subscriptions.push(vscode_1.window.onDidChangeActiveTextEditor(() => this.updateInputActive()));
        this.subscriptions.push(vscode_1.commands.registerTextEditorCommand('lean.input.convert', (editor, edit) => {
            const handler = this.handlers.get(editor);
            if (handler) {
                handler.convertRange();
            }
        }));
        this.subscriptions.push(vscode_1.workspace.onDidChangeConfiguration(() => {
            this.leader = inputModeLeader();
            this.enabled = inputModeEnabled();
            this.languages = inputModeLanguages();
            this.customTranslations = inputModeCustomTranslations();
            this.allTranslations = Object.assign(Object.assign({}, this.translations), this.customTranslations);
        }));
    }
    setInputActive(isActive) {
        vscode_1.commands.executeCommand('setContext', 'lean.input.isActive', isActive);
    }
    get active() {
        const handler = this.handlers.get(vscode_1.window.activeTextEditor);
        return handler && !!handler.range;
    }
    updateInputActive() {
        this.setInputActive(this.active);
    }
    findReplacement(typedAbbrev) {
        if (typedAbbrev === '') {
            return undefined;
        }
        if (this.allTranslations[typedAbbrev]) {
            return this.allTranslations[typedAbbrev];
        }
        let shortestExtension = null;
        for (const abbrev in this.allTranslations) {
            if (abbrev.startsWith(typedAbbrev) && (!shortestExtension || abbrev.length < shortestExtension.length)) {
                shortestExtension = abbrev;
            }
        }
        if (shortestExtension) {
            return this.allTranslations[shortestExtension];
        }
        else if (typedAbbrev) {
            const prefixReplacement = this.findReplacement(typedAbbrev.slice(0, typedAbbrev.length - 1));
            if (prefixReplacement) {
                return prefixReplacement + typedAbbrev.slice(typedAbbrev.length - 1);
            }
        }
        return null;
    }
    isSupportedFile(document) {
        return !!vscode_1.languages.match(this.languages, document);
    }
    onChanged(ev) {
        const editor = vscode_1.window.activeTextEditor;
        if (editor.document !== ev.document) {
            return;
        } // change happened in active editor
        if (!this.isSupportedFile(ev.document)) {
            return;
        } // Not a supported file
        if (!this.handlers.has(editor)) {
            this.handlers.set(editor, new TextEditorAbbrevHandler(editor, this));
        }
        this.handlers.get(editor).onChanged(ev);
    }
    onSelectionChanged(ev) {
        const editor = vscode_1.window.activeTextEditor;
        if (editor !== ev.textEditor) {
            return;
        } // change happened in active editor
        if (!this.isSupportedFile(editor.document)) {
            return;
        } // Lean file
        if (this.handlers.has(editor)) {
            this.handlers.get(editor).onSelectionChanged(ev);
        }
    }
    dispose() {
        this.decorationType.dispose();
        for (const s of this.subscriptions) {
            s.dispose();
        }
    }
}
exports.LeanInputAbbreviator = LeanInputAbbreviator;
//# sourceMappingURL=input.js.map